#ifndef HWACCESS_INCLUDED
#define HWACCESS_INCLUDED



#ifndef TYPES_INCLUDED
    #error "Needs types(Types.inc)."
    //#include "Types.inc"
#endif

#ifndef GLOBALS_INCLUDED
    #error "Needs global types and variables(Globals.inc)."
    //#include "Globals.inc"
#endif

#ifndef HWCORE_INCLUDED
    #error "Needs hardware core(HWCore.inc)."
    //#include "HWCore.inc"
#endif

#ifndef ROBDEBUG_INCLUDED
    #error "Needs robust debugging(RobDebug.inc)."
    //#include "RobDebug.inc"
#endif

void PrinterPortByteWrite(U8 Value)
{   PDL("PrintPortByteWrite(U8 Value): Control Printer Port Digital Outputs");
    Deb("Do Printer Pins Default to output?");
    outportb(PRINTER_PORT_ADDRESS, Value);
}

void PCIDetect(U16 TargetVID, U16 TargetDID)
{
    U16 Bus, Device;
    union
    {
        struct
        {
            unsigned char Addr;
            unsigned char Function: 3;
            unsigned char Device: 5;
            unsigned char Bus;
            unsigned char Header;
        } P;
        unsigned long Trans;
    } Trans;

    DIO96Base = 0;

    Trans.P.Header = 0x80;
    Trans.P.Addr = 0;
    Trans.P.Function = 0;
    for ( Bus = 0; Bus < 256; ++Bus )
    {
        Trans.P.Bus = Bus;
        U16 DevicesOnBus = 0;
        for ( Device = 0; Device < 32; ++Device )
        {
            Trans.P.Device = Device;
            U16 ID;
            OutPortDW(0xCF8, Trans.Trans);
            ID = InPort(0xCFC);
            if ( ID == 0xFFFF ) continue;
            ++DevicesOnBus;
            if ( ID != TargetVID ) continue;
            ID = InPort(0xCFE);
            if ( ID != TargetDID ) continue;

            OutPortDW(0xCF8, Trans.Trans | 0x18);
            DIO96Base = InPortDW(0xCFC) & 0xFFF8;
            break;
        }
        if ( DIO96Base ) break;
        if ( DevicesOnBus == 0 ) break;
    }

}

void GetMasterTick(TMasterTime & Target)
{
    __emit__(0x0F, 0x31);
    Target.Lo32 = _EAX;
    Target.Hi32 = _EDX;
}

void DetectClockHz(void)
{
    //delay(1);
    TMasterTime Tick;
    GetMasterTick(Tick);
    delay(1000);
    GetMasterTick(ClockHz);
    ClockHz -= Tick;
    //ClockHz *= 1.0L/5.0L;
}

TBool ReadHex32(char * Src, U32 & Dest)
{
	for ( int I = 0; I < 8; ++I )
	{
		if ( (*Src >= '0') && (*Src <= '9') )
			Dest = (Dest << 4) | (*Src - '0');
		if ( (*Src >= 'A') && (*Src <= 'F') )
			Dest = (Dest << 4) | (*Src - ('A' - 10));
		if ( (*Src >= 'a') && (*Src <= 'f') )
			Dest = (Dest << 4) | (*Src - ('a' - 10));
		else
			return False
		;
		++Src;
	}
	return True;
}

void LoadClockHz(void)
{
	TMasterTime TempHz = { 0, 0 };
	ClockHz = TempHz;
	FILE *F = fopen("Clock.txt", "rb");
	if ( F == 0 ) return;

	char Buf[16];
	int L = fread(Buf, 1, 16, F);
	if ( L != 16 ) return;

	if ( ! ReadHex32(Buf, TempHz.Hi32) ) return;
	if ( ! ReadHex32(Buf+8, TempHz.Lo32) ) return;
	ClockHz = TempHz;
}

void InitClockHz(void)
{
    LoadClockHz();
    if ( (ClockHz.Lo32 | ClockHz.Hi32) == 0 ) DetectClockHz();

    double SecsPerMasterTick = 1.0L / MasterTickToFloat(ClockHz);
    USecsPerMasterTick = SecsPerMasterTick * 1000000.0L;
    MinsPerMasterTick = SecsPerMasterTick / 60.0L;
    MSecsPerMasterTick = SecsPerMasterTick * 1000.0L;
    double fBigTicksPerKeepalive = MasterTickToFloat(ClockHz) * 45.0L / (65536.0L*65536.0L);
    BigTicksPerKeepalive = fBigTicksPerKeepalive;

    TMasterTime Tick;
    GetMasterTick(Tick);
    LastKeepaliveBigTick = Tick.Hi32;
}

int __fastcall GotData(void)
{
    volatile U8 Status = InPortB(DIO96Base + 7);
    Status &= 0x80;
    return Status;
}

void __fastcall GetDatum(void)                  // Take Data.  Must be FAST.  Abuses (makes optimum use of) structures to pump 32-bit data around fast.
{
    PDL("void GetOtherDatum(void): Gets one data point off the card for one X I-Beams and one Y I-Beam.");
    pU32 pData = pU32(&MostRecentData.Datum.X[0]);
    *(pData++) = InPortDW(DIO96Base);
    *(pData++) = InPortDW(DIO96Base);
    *(pData  ) = InPortDW(DIO96Base);
    GetMasterTick(MostRecentData.Datum.SampleTime);
}

void SetMask(void)
{
//    pU32 pData = Mask.By32;
    OutPortDW(DIO96Base + 0x10, Mask.By32[0]);// *(pData++));
    OutPortDW(DIO96Base + 0x14, Mask.By32[1]);// *(pData++));
    OutPortDW(DIO96Base + 0x18, Mask.By32[2]);// *(pData++));
}

void ClearMask()
{
    ForInt(i,0,4)
        Mask.By32[i]=0L;
    EndFor;
    SetMask();
}

TMasterTime BenchTime;

void __fastcall BenchOn(void)
{
    GetMasterTick(BenchTime);
}

void __fastcall BenchOff(TBool Wait)
{
    TMasterTime BenchDone;
    GetMasterTick(BenchDone);
    double dT = MasterTickToFloat(BenchDone - BenchTime);
    dT *= USecsPerMasterTick;
    char msg[80];
    sprintf(msg,"Benchmark: %lfus", dT);
      puts("");
      puts(msg);
    if (Wait)
      getch();
}

void __fastcall BenchOff(char* msg, TBool Wait)
{
    BenchOff(Wait);
    puts(msg);
}

float *BenchAryFloat;
I16 BenchAryFloatIndex;

void InitBenchLog(void)
{
    BenchAryFloatIndex = 0;
    BenchAryFloat = (float *) malloc(10000*sizeof(float));
}

void DeInitBenchLog(void)
{
    if (BenchAryFloat)
    {
        free(BenchAryFloat);
        BenchAryFloat = 0L;
    }
}

void __fastcall BenchLog(void)
{
    TMasterTime BenchDone;
    GetMasterTick(BenchDone);
    float dT = MasterTickToFloat(BenchDone - BenchTime);
    dT *= USecsPerMasterTick;
    BenchAryFloat[BenchAryFloatIndex] = dT;
    //BenchAryFloatIndex++;
}



#endif