#ifndef TYPES_INCLUDED
#define TYPES_INCLUDED



typedef unsigned char     U8 , *pU8 ;           // bytes or keystrokes or string elements
typedef unsigned int      U16, *pU16;           // words
typedef unsigned long int U32, *pU32;           // double-words
typedef signed char       I8 , *pI8 ;           // short integers
typedef        int        I16, *pI16;           // 16-bit integers
typedef        long int   I32, *pI32;           // 32-bit integers

typedef enum { False = 0, True = 1 } TBool;      // Boolean
#define FALSE False
#define TRUE True



typedef enum
{
    StatusOK,
    StatusUserAbort, // Keystroke abort code, utter and urgent
    StatusExitCommand, // TCP Command to Exit
    StatusFatalError,
    StatusErrorRecoveryFailed,
    StatusNoHardware,
    StatusTwoCardsNoLX,
    StatusNopeJustFlicker,
    StatusCount
} TStatus;

typedef struct
{
  U32 Lo32, Hi32;
} TMasterTime;

double MasterTickToFloat(TMasterTime & Target)
{
    double Result = Target.Hi32;
    Result *= 65536.0L * 65536.0L;
    Result += Target.Lo32;
    return Result;
}

TMasterTime operator-(TMasterTime A, TMasterTime & B)
{
    if ( B.Lo32 > A.Lo32 ) --A.Hi32;
    A.Lo32 -= B.Lo32;
    A.Hi32 -= B.Hi32;
    return A;
}

void operator-=(TMasterTime & A, TMasterTime & B)
{
    if ( B.Lo32 > A.Lo32 ) --A.Hi32;
    A.Lo32 -= B.Lo32;
    A.Hi32 -= B.Hi32;
}

double operator*(TMasterTime & A, double B)
{
    return MasterTickToFloat(A) * B;
}



typedef U8 TIBeam[6];                           // The data from a single sensor I-Beam

typedef struct
{
    TMasterTime SampleTime;
    TIBeam X, Y, LX;                            // "X" is also known as "Vertical lasers", "Y" can be called "Horizontal lasers", and LX is "the
                                                //      second set of vertical lasers associated with the pitcher's grid"
} TDatum, *pTDatum;                                       // TDatum is a snapshot in time of the data from one lane's single grid (one square of data), with


#endif