#ifndef GLOBALS_INCLUDED
#define GLOBALS_INCLUDED



#ifndef TYPES_INCLUDED
    #error "Needs types(Types.inc)."
    //#include "Types.inc"
#endif


double SyncOffsetTime = 0;
double LastPitchTime = 0;

const char HexChar[16] = "0123456789ABCDEF";
const char *StatusToStr[StatusCount] = {"OK", "User Abort!", "Exit Command", "Fatal Error", "Error Recovery Failed", "No Hardware", "Nope, Just Flicker"};

#define cmdMASK 0x4B53414D
#define cmdSYNC 0x434E5953
#define cmdPROG 0x474F5250
#define cmdMORE 0x45524F4D
#define cmdDONE 0x454E4F44
#define cmdBOOT 0x544F4F42

typedef union
{
    struct
    {
        TIBeam Beam[3];
        U16 PadForIOWrite;
    }ByBeam;
    U32 By32[5];
} TMask;

typedef struct
{
    TDatum Datum;
    U16 PadForIORead;                           // This U16 allows us to use a inportDW twice to grab the data, and simply ignore the extra bits
} TMostRecentData;

typedef struct
{
    I16 Counter[3][48];                         // Each sensor is associated with a counter that increments-while-blocked.  This is the storage for
                                                //  all three I-Beams worth of data
} TCounters, *pTCounters;

typedef struct
{
    double StartTime, EndTime;
    U16 BitIndex;
} TTimestamps, *pTTimestamps;

typedef enum
{
    byLength, byIndex
} TSortBy;

typedef enum
{
    sInit,              //"0", consider moving out of state machine
    sBeginState,        //not on paper.  Might be useful to have a state that is always the target of "my work is done, begin anew" state-loop-path-endings
    sCheckAndParseMsg,  //1 - these numbers are from the pencil drawing state diagram.  formalize and friendly name fix. -- J2H
    sCheckAndReadSensors, //2 & 3
    //sTransformData,     //4
    sImplementCounters, //5
    //sCheckOtherHoriz,   //6
    //sSaveData,          //7
    //sUnusedState,       //8
    sNoDark,            //9
    //sCheckFlicker,      //10 //currently does two things. consider diverting into a sub-state-machine instead (not_my_horiz_pitch condition)
    sParseData,         //11
    sQueueDataForXmit,  //12
    //sNotMyPitch,        //13
    //sIfOtherHorizTrue,  //14
    sQueueErrorForXmit, //15
    sDeInit,            // consider moving out of state machine
    NumberOfStates,      // how amazingly useful --J2H
    sInvalidState = -1
} TState;

char *GetStateName[] =
{
    "Init",
    "BeginState",
    "CheckAndParseMsg",
    "CheckSensors",
    "ReadSensors",
    //"TransformData",
    "ImplementCounters",
    //"CheckOtherHoriz",
    //"SaveData",
    //"UnusedState",
    "NoDark",
    //"CheckFlicker",
    "ParseData",
    "QueueDataForXmit",
    //"NotMyPitch",
    //"IfOtherHorizTrue",
    "QueueErrorForXmit",
    "DeInit",
};


I16 GridNumber = 0;


I16 TERMINATE = 0;  // This global variable serves application.terminate purposes.  Normally called "Done" by J2H

TMasterTime ClockHz;
double USecsPerMasterTick, MinsPerMasterTick, MSecsPerMasterTick;
U32 BigTicksPerKeepalive, LastKeepaliveBigTick;

//#define MPH_TO_IPU (0.0000176L)
//#define MAX_MPH 120.0
//#define MIN_BALL_RAD 1.40L
//#define MIN_REAL_LONGEST (MIN_BALL_RAD * 2 / (MAX_MPH * MPH_TO_IPU))
#define MIN_REAL_LONGEST 500
#define MAX_REAL_LASERS_INTERCEPTED 3
TMostRecentData MostRecentData;                 // Used to read data into, and to increment counters
#define RawDataCount 2048
I16 NextRawData = 0;                            // "Depth of array" indicates how much data is currently in the RawData struct
TDatum RawData[RawDataCount];                            // Used to store data until ParsePitch happens
U32 Sentinel_DEBUG = 0xAA558907L;
I32 ThisFrameID = 0, LastFrameID = -1;
U16 DIO96Base[2];                               // Base Addresses of the (up to) two I/O cards in the system.  Stuffed by PCI plug-and-play
void __fastcall (*GetDatum)(void);
int __fastcall (*GotData)(void);

TMask BlockedDark, BlockedLight, Flicker, Mask;

int min(int a, int b)
{
    return (a<b)?a:b;
}

typedef struct
{
    U32 Length;
    U8 Data[256];
} TMsgPart;

typedef TMsgPart TMsgHeader;
typedef TMsgPart TMsgPitchData;
typedef TMsgPart TMsgErrorData;

TMsgHeader MsgHeader;
TMsgPitchData MsgPitchData;
TMsgErrorData MsgErrorData;
char bufMsgToSend[2000];
TBool bMsgQueuedForSend;

void ClearDatums(void)
{
    NextRawData = 0;
    ++ThisFrameID;
}

void AddStringToMsgPart(TMsgPart & Target, char * StrBuf)
{
    PDL("void AddStringToMsgPart(TMsgPart & Target, char * StrBuf): adds a string to the message header or error data. If the string contains a comma, that just separates it into more strings.");
    U16 L = strlen(StrBuf);
    if ( Target.Length ) Target.Data[Target.Length++] = ',';
    memcpy(Target.Data + Target.Length, StrBuf, L);
    Target.Length += L;
}

TStatus QueueMsgToSend(TMsgHeader *head, TMsgPitchData *pitch, TMsgErrorData *err)
{
    U32 len = 4;
    if (head)
    {
        memcpy(bufMsgToSend + len, head, (*(U32 *)head) + 4);
        len += *(U32 *)head + 4;
    }
    else
    {
        memset( bufMsgToSend + len, 0, 4 );
        len += 4;
    }

    if (pitch)
    {
        memcpy(bufMsgToSend + len, pitch, (*(U32 *)pitch) + 4);
        len += *(U32 *)pitch + 4;
    }
    else
    {
        memset( bufMsgToSend + len, 0, 4 );
        len += 4;
    }

    if (err)
    {
        memcpy(bufMsgToSend + len, err, (*(U32 *)err) + 4);
        len += *(U32 *)err + 4;
    }
    else
    {
        memset( bufMsgToSend + len, 0, 4 );
        len += 4;
    }

    *(U32 *)bufMsgToSend = (len-4);

    if (len)
        bMsgQueuedForSend = TRUE;
    else
        bMsgQueuedForSend = FALSE;

    return StatusOK;
}

int ClearMsgToSend()
{
    bMsgQueuedForSend = FALSE;
    memset(bufMsgToSend,0,sizeof(bufMsgToSend));
    return 0;
}

TBool MsgToSend()  //used like "if (MsgToSend()) sock_write(s, bufMsgToSend, *(U32 *)bufMsgToSend);"
{
    return bMsgQueuedForSend;
}

TCounters BlockCount, FlickerCount;

void CountFlicker(U16 BitIndex)
{
    char Buf[3] = "\0\0\0";
    Buf[0] = '0' + BitIndex / 10;
    Buf[1] = '0' + BitIndex % 10;
    Deb("Counting flicker on this bit:");
    Deb(Buf);
    ++FlickerCount.Counter[BitIndex / 48][BitIndex % 48];
}

void ClearBlockCounters(void)
{
    memset(&BlockCount, 0, sizeof(BlockCount));
    /*
    ForInt(IBeam, 0, 2)
        ForInt(IByte, 0, 5)
            BlockCount.Counter[IBeam][IByte] = 0;
        EndFor;
    EndFor;
    */
}

void ClearFlickerCounters(void)
{
    memset(&FlickerCount, 0, sizeof(FlickerCount));
}

TBool bInPitch;

void SetInPitch(TBool is)
{
    bInPitch = is;
}

TBool InPitch(void)
{
    return bInPitch;
}

char *FlickersToStr(void)
{
    static char Buf[256];
    char *pBuf = Buf;
    PDL("char *FlickersToStr(void): Convert flicker counters to a static string.");

    ForIntDown(BitIndex, 47, 0)
        *(pBuf++) = HexChar[min(0xF, FlickerCount.Counter[0][BitIndex])];
    EndFor;
    *(pBuf++) = ',';
    ForIntDown(BitIndex, 47, 0)
        *(pBuf++) = HexChar[min(0xF, FlickerCount.Counter[1][BitIndex])];
    EndFor;
    *(pBuf++) = ',';
    if ( DIO96Base[1] )
    {
        ForIntDown(BitIndex, 47, 0)
            *(pBuf++) = HexChar[min(0xF, FlickerCount.Counter[2][BitIndex])];
        EndFor;
    }
    *pBuf = 0;
    return Buf;
}

void ComplainUntilKBHit(I32 HighTone)
{
    ++HighTone;
    do
    {
        for ( I32 I = HighTone; I >= 0; --I )
        {
            int Pitch = 100 + (double(I) / double(HighTone) * 1900);
            sound(Pitch); delay(300);
        }
        nosound(); delay(300);
    } while ( ! kbhit() );
}

void CheckFrameID(char *Wot)
{
    if ( LastFrameID == ThisFrameID )
    {
        printf("Debug frame ID duplicate on %s!\n", Wot);
        ComplainUntilKBHit(2);
    }
    LastFrameID = ThisFrameID;
}

void BuildMaskFlickerErrorData(void)    //build response to MASKWHAT (MASK?)
{
    char Buf[256];
    char *pBuf = Buf;
    int HighBeam = DIO96Base[1] ? 2 : 1;
    ForInt(IBeam, 0, HighBeam)
        ForIntDown(IByte, 5, 0)
            *(pBuf++) = HexChar[(Mask.ByBeam.Beam[IBeam][IByte]) >> 4];
            *(pBuf++) = HexChar[(Mask.ByBeam.Beam[IBeam][IByte]) & 0xF];
        EndFor;
        if ( IBeam != 2 ) *(pBuf++) = ',';
    EndFor;
    *(pBuf++) = 0;

    MsgErrorData.Length = 0;
//    AddStringToMsgPart(MsgErrorData, "=");
    AddStringToMsgPart(MsgErrorData, Buf);
    AddStringToMsgPart(MsgErrorData, FlickersToStr());
}

void BuildBlockFlickerErrorData(void)
{
    TDatum DataForMask;

    DataForMask = RawData[RawDataCount - 1];
    ForInt(I, RawDataCount / 2, RawDataCount - 2)
        ForInt(IByte, 0, 5)
            DataForMask.X[IByte] |= RawData[I].X[IByte];
            DataForMask.Y[IByte] |= RawData[I].Y[IByte];
            DataForMask.LX[IByte] |= RawData[I].LX[IByte];
        EndFor;
    EndFor;

    char Buf[256];
    char *pBuf;
    ForInt(IByte, 0, 5)
        DataForMask.X[IByte] ^= 0xFF;
        DataForMask.Y[IByte] ^= 0xFF;
        DataForMask.LX[IByte] ^= 0xFF;//invert block masks for active"high"
    EndFor;

    pBuf = Buf;
    ForIntDown(IByte, 5, 0)
        Mask.ByBeam.Beam[0][IByte] |= DataForMask.X[IByte];
        *(pBuf++) = HexChar[Mask.ByBeam.Beam[0][IByte] >> 4];
        *(pBuf++) = HexChar[Mask.ByBeam.Beam[0][IByte] & 0xF];
    EndFor;
    *(pBuf++) = ',';
    ForIntDown(IByte, 5, 0)
        Mask.ByBeam.Beam[1][IByte] |= DataForMask.Y[IByte];
        *(pBuf++) = HexChar[Mask.ByBeam.Beam[1][IByte] >> 4];
        *(pBuf++) = HexChar[Mask.ByBeam.Beam[1][IByte] & 0xF];
    EndFor;
    *(pBuf++) = ',';
    if ( DIO96Base[1] )
    {
        ForIntDown(IByte, 5, 0)
            Mask.ByBeam.Beam[2][IByte] |= DataForMask.LX[IByte];
            *(pBuf++) = HexChar[Mask.ByBeam.Beam[2][IByte] >> 4];
            *(pBuf++) = HexChar[Mask.ByBeam.Beam[2][IByte] & 0xF];
        EndFor;
    }
    *(pBuf++) = 0;


    MsgErrorData.Length = 0;
    AddStringToMsgPart(MsgErrorData, Buf);
    AddStringToMsgPart(MsgErrorData, FlickersToStr());
}

void ReportBlocked(void)
{
    CheckFrameID("block");

    BuildBlockFlickerErrorData();

    QueueMsgToSend(&MsgHeader, NULL, &MsgErrorData);
}

/*
Conceptual "member function" of the data store array.  Stuffs an element.  We've decided not to
     implement an entire complete class structure as only two functions use the array's contents,
     this one and the ParsePitch, which uses it in random-access mode for now.
*/
TBool AddDatum(void)
{
    if ( NextRawData < RawDataCount )
    {
        RawData[NextRawData] = MostRecentData.Datum;
        ++NextRawData;
        return True;
    }
    else
    {
        if ( NextRawData == RawDataCount )
        {
            Rem("RawData array is full, must be a blocked sensor.");
            Deb("Block Detected");
            ReportBlocked();
            Deb("Block Finished Parsing and in queue");
            ++NextRawData;
        }
        SetInPitch(FALSE);
        return False;
    }
}

#define ResultsCount 144
TTimestamps XResults[ResultsCount], YResults[ResultsCount], LXResults[ResultsCount];
I16 NextXResult = 0, NextYResult = 0, NextLXResult = 0;

pTTimestamps GetMyTimestamps(U16 BitIndex)
{
    TTimestamps * Results;
    I16 * NextResult;

    if ( BitIndex < 48 )
    {
        Results = XResults;
        NextResult = &NextXResult;
    }
    else if ( BitIndex < 48*2 )
    {
        Results = YResults;
        NextResult = &NextYResult;
    }
    else
    {
        Results = LXResults;
        NextResult = &NextLXResult;
    }

    ForInt(I, 0, *NextResult - 1)
        if ( Results[I].BitIndex == BitIndex )
            return &(Results[I])
        ;
    EndFor;
    if ( *NextResult >= ResultsCount )
    {
        Deb("More intercepts to timestamp than available timestamp slots.");
        sound(1000);
        delay(1000);
        nosound();
    }
    else
    {
        Results[*NextResult].BitIndex = BitIndex;
        return &(Results[(*NextResult)++]);
    }
    return 0;
}


void ClearTimestamps(void)
{
    NextXResult = 0;
    NextYResult = 0;
    NextLXResult = 0;
}

void SortTimestamps(TTimestamps Timestamps[], U16 NumTimestamps, TSortBy ByWhat)
{
    switch ( ByWhat )
    {
        case byIndex:
            ForIntDown(I, NumTimestamps - 2, 0)
                ForInt(J, 0, I)
                    if ( Timestamps[J].BitIndex > Timestamps[J + 1].BitIndex )
                    {
                        TTimestamps Buf;
                        Buf = Timestamps[J];
                        Timestamps[J] = Timestamps[J + 1];
                        Timestamps[J + 1] = Buf;
                    }
                EndFor;
            EndFor;
        break;
        case byLength:
            ForIntDown(I, NumTimestamps - 2, 0)
                ForInt(J, 0, I)
                    if ( Timestamps[J].EndTime - Timestamps[J].StartTime < Timestamps[J + 1].EndTime - Timestamps[J + 1].StartTime )
                    {
                        TTimestamps Buf;
                        Buf = Timestamps[J];
                        Timestamps[J] = Timestamps[J + 1];
                        Timestamps[J + 1] = Buf;
                    }
                EndFor;
            EndFor;
        break;
    }
}

TState cur_State = sInit, nxt_State = sInvalidState, prv_States[NumberOfStates];
U16 prv_State_Index = 0, prv_State_Depth = 0;



TBool bMessageWaiting;



#endif