/*

Some definition of terms.

Lane: The complete set of sensors for any given pitcher/catcher pair.
Plane: rare; the intersection of a plane perpendicular to the line(s) between pitcher(s) and the appropriate catcher(s).
Truss: the group of sensors built into a device on a single Plane
Grid: the group of sensors associated with a single Lane of a single Truss.

I-Beam: A single 48-laser or 48-sensor equipped, extruded aluminum, I-Beam.  One mounted on the vertical of each truss, one on the horizontal of
    each lane of each truss, and a third mounted also on the horizontal, but only on the front truss (closest to pitcher).  Specifically the sensor
    equipped I-Beams, as we don't usually care about the laser-equipped I-Beams from software.

Data Acquisition Computer: A PC system intended to acquire data from a single grid of sensors.  Intended to be ETX/embedded.
Log In Computer: A single computer intended to be located at the first truss (nearest the pitcher) with a touch-screen interface to log-in.
Display Computer: A computer+monitor per lane that provides feedback to the pitcher for that lane.  Big Monitor, as it is designed to be mounted
    12 feet away from pitcher on the top of the first truss
Boss-Comp, or Calculation Computer: A single server-class system which marshalls all other systems across an ethernet star network.

s: The socket for TCP/IP

*/

//**************************************************************************************************************************************************
//  BEGIN VERSION_INFO Initialization

#define VerStr "Rev 0.49"
#define VerStrHTML "Rev%200.49"
#define REVISION    0x0049

#define byCOMPANY "ACCES I%2FO Products, Inc"
#define forCOMPANY "for OnePitch (c) 2006-2009"
#define MODEL "Model LIVEMODE"
#define PRINTER_PORT_ADDRESS 0x378

#define HostName "10.1.1.3"
#define DebugHostName "10.1.1.207"
#define IP_PORT 4242

//  END VERSION_INFO Initialization
//**************************************************************************************************************************************************

//**************************************************************************************************************************************************
// Begin Includes
//

#pragma inline

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <process.h>

extern "C"
{
    #include <tcp.h>
}


#include "Types.inc"
#include "HWCore.inc"
#include "RobDebug.inc"
#include "Globals.inc"
#include "StrConvs.inc"
#include "HWAccess.inc"

//
//  END Includes
//**************************************************************************************************************************************************

//**************************************************************************************************************************************************
// Begin Not In A Region
//

tcp_Socket sock;
tcp_Socket *s = &sock;

char buf[2048]; //not to be longer than one enet packet (1344?) -- J2H
char bufMsgReceived[2000];
//FILE *ProgFile = 0;

I16 DoStateDeInit(void);

//
// End Not In A Region
//**************************************************************************************************************************************************

//*********************************
// Begin State Machine Tools
//

TState SetNextState(TState NewState = sBeginState)
{
    if ( nxt_State != sInvalidState )
    {
        char Msg[256];
        //sprintf(Msg, "The program overwrote nxt_State; was %s now %s.", StateToName[nxt_State], StateToName[NewState]);
        //Rem(Msg);
    }

    nxt_State = NewState;
    return sInit;
}

TState GetCurrentState(void)
{
    return cur_State;
}

U16 GetPreviousStateDepth()
{
    return prv_State_Depth;
}

TState GetPreviousState(U16 Depth = 0)
{
    I16 ThatState = prv_State_Index - Depth;
    while ( ThatState < 0 ) ThatState += NumberOfStates;
    return prv_States[ThatState];
}

void ChangeState(void)
{
    if (Sentinel_DEBUG !=0xAA558907L)
    {
      Sentinel_DEBUG = 0xAA558907L;
      cprintf("ARRAY CORRUPTION: State = %s\r\n",GetStateName[cur_State]);
      ComplainUntilKBHit(1);
    }

    if ( nxt_State == sInvalidState )
        Deb("ChangeState() was called without setting next state! Ignoring state change.");
    else
    {
        prv_State_Index = (prv_State_Index + 1) % NumberOfStates;
        ++prv_State_Depth;
          if ( prv_State_Depth > NumberOfStates ) prv_State_Depth = NumberOfStates;
          prv_States[prv_State_Index] = cur_State;
          cur_State = nxt_State;
          nxt_State = sInvalidState;
          #if (SOURCE_VERBOSITY > VERBOSITY_S_REM)
              fputs("\n", stderr);
          #endif

    }
    if (TERMINATE) cur_State = sDeInit;
    //Deb(GetStateName[cur_State]);
}

//
// End State Machine Tools
//*********************************



//**************************************************************************************************************************************************
// Begin Ethernet Crap
//

#pragma argsused
U16 _fastcall MotorolaWord(U16 *W)
{
    asm xchg AH, AL;
    return _AX;
}

#pragma argsused
U32 _fastcall MotorolaDWord(U32 *D)
{
    asm {
        rol EAX, 16;
        mov AX, DX;
        xchg AH, AL;
    }
    return _EAX;
}

void SetInMessage(U8 *buf, size_t len)
{
    memcpy(bufMsgReceived, buf, min(len, (int)sizeof(bufMsgReceived)));
    bufMsgReceived[min(len, (int)sizeof(bufMsgReceived)-1)] = 0;
}

void SetMessageWaiting(TBool is)
{
    bMessageWaiting = is;
}

TBool MessageWaiting(void)
{
    return bMessageWaiting;
}

#pragma argsused
int HandleTCP(void *s, U8 *data, int len, tcp_PseudoHeader *pseudohdr, void *protohdr)
{
    if (!len)
        return len;

    if ( MessageWaiting() )
    {
        Deb("Busy");
        return 0; //(if "already" a message waiting
    }

    if (len >= 4)
    {
        //packet data indicates total length, and len indicates amount received so far
        if (*(U32 *)data + 4 > 0x000000FF)
        {
            SetNextState(sInit);
            return len;
        }

        if (len < *(U32 *)data + 4 )
        {
            //printf("Got partial packet HandleTCP len=%d, packlen=%ld",len,*(U32 *)data);
            return 0;//not received everything yet, don't parse anything
        }
        else //if (len >= *(U32 *)data + 4)  //we received all of a packet
        {
            PDL("Handle Incoming TCP Data on S connection");
            PDL("In general, just copy the packet's data into global message holder and set 'MessageWaiting'");
            SetInMessage( data, *(U32 *)data + 4 );//copy packet into incoming message queue
            SetMessageWaiting( TRUE );

            return *(U32 *)data + 4;  //only return this packet's length, so we get called with next packet separately.
        }
    }
    return 0;
}

void PerformDebugTCP(void)
{
    int Status = StatusOK;

    U32 host, Len;
    host = resolve(DebugHostName);

    if (!tcp_open( s, 0, host, IP_PORT, HandleTCP ))
    {
        Deb("Debug_tcp_open failed, not sure why.");
        return;
    }
             sock_wait_established( s, sock_delay, NULL, &Status );

    //Send default debug message.
    MsgErrorData.Length = 0;
    AddStringToMsgPart(MsgErrorData, "something is wrong with my Boss-Comp");
    QueueMsgToSend(&MsgHeader, NULL, &MsgErrorData);
    sock_write(s, bufMsgToSend, *(U32 *)bufMsgToSend+4);
    ClearMsgToSend();

    return;

sock_err:
    Deb("TCP/IP Error during connect to Debug Boss-Comp");
}


void Reboot(void)
{
    //asm  JMP 0xFFFF:0x0000;//note, JMP doesn't work, AJMP, LJMP etc all fail.  needs to be "FAR ABSOLUTE JUMP" but we don't know how.
    DoStateDeInit();
//    system("SmartDrv /C");
    system("Reboot.com");
    puts("Should Have Rebooted, not gotten this far: aborting.");
    exit(252);
}

void Connect(void)
{
    int Status = StatusOK;

    PDL("void Connect(void);  Open a global communication channel with Boss-comp over TCP/IP.");

    U32 host;
    TMasterTime StartTime, Naui;
    double MinutesTrying;
    TBool bFiveMinuteCheck = False;

    host = resolve(HostName);

    GetMasterTick(StartTime);
    while ( ! tcp_open(s, 0, host, IP_PORT, HandleTCP) )
    {
sock_err:
        Deb("Trying..\r\n");
        GetMasterTick(Naui);
        MinutesTrying = MasterTickToFloat(Naui - StartTime) * MinsPerMasterTick;
        if ( MinutesTrying >= 60.0 ) Reboot();
        if ( (MinutesTrying >= 5.0) && ! bFiveMinuteCheck )
        {
            PerformDebugTCP();
            bFiveMinuteCheck = True;
        }

    }
    sock_mode(s,TCP_MODE_BINARY);

    GetMasterTick(StartTime);
    bFiveMinuteCheck = False;
    sock_wait_established( s, sock_delay, NULL, &Status );
    Print("\r\nConnected\r\n");

    return;

}

//
// End Ethernet Crap
//**************************************************************************************************************************************************

//**************************************************************************************************************************************************
// Begin State Logic
//

I16 DoStateInit(void)
{
    int Status = StatusOK;

    PDL("State: Init.  Consider moving out of state machine.");
    PDL("   Perform all initialization functionality that needs to be done only once per program run.");
    PDL("   one-time de-init will be performed by State: DeInit.");
    PDL("   This state must detect how many cards installed, configure the address arrays, perform handshaking with enet boss-comp about identity, synch time, wait for go...");

    PCIDetect(0x494F, 0x0C69);
    if ( ! DIO96Base[0] )
    {
        Print("No hardware found, exiting.\r\n");
        TERMINATE = StatusNoHardware;
    }
    if ( DIO96Base[1] )
    {
        U32 LastData;

        PDL("We have two cards.");
        Rem("Set which GetDatum and GotData functions to use.");
        GetDatum = GetFrontDatum;
        GotData = GotFrontData;
        Rem("Test and swap cards if needed.");
        InPortDW(DIO96Base[1]);
        InPortDW(DIO96Base[1]);
        LastData = InPortDW(DIO96Base[1]);
        if (( LastData & 0x00FF0000 ) == 0x00FF0000 )
        {
            LastData = DIO96Base[0];
            DIO96Base[0] = DIO96Base[1];
            DIO96Base[1] = LastData;
        }
        else //we have two cards, but the second one wasn't LX plugged
        {
            InPortDW(DIO96Base[0]);
            InPortDW(DIO96Base[0]);
            LastData = InPortDW(DIO96Base[0]);
            if ( LastData & 0x00FF0000 )
            {
              Rem("Error, NEITHER card indicates it is LX");
              Deb("FATAL ERROR, NEITHER CARD IS LX CODED (wrap-plug missing?).");
              TERMINATE = StatusTwoCardsNoLX;
              ComplainUntilKBHit(0);
              Status = StatusNoHardware;
            }
        }
    }
    else
    {
        PDL("We have one card.");
        Rem("Set which GetDatum and GotData functions to use.");
        GetDatum = GetOtherDatum;
        GotData = GotOtherData;
    }
    if ( TERMINATE ) return TERMINATE;

    GetDatum();

    ClearMask();

    InitClockHz();
    {
        char msg[256];
        if (DIO96Base[1])
            sprintf(msg,"\nBase[0]=%04X Base[1]=%04X\r\n"
                        "CPU MHz = %5.3lf\r\n",
                        DIO96Base[0],
                        DIO96Base[1],
                        (double)(MasterTickToFloat(ClockHz)/1000000.0L));
        else
            sprintf(msg,"\nBase[0]=%04X\r\n"
                        "CPU MHz = %5.3lf\r\n",
                        DIO96Base[0],
                        (double)(MasterTickToFloat(ClockHz)/1000000.0L));
        Print(msg);
    }

    PDL("Begin Ethernet Init");
    MsgHeader.Length = 0;
    AddStringToMsgPart(MsgHeader, "OnePitch%20Product");
    AddStringToMsgPart(MsgHeader, "by%20ACCES%20I%2FO%20Products%20Inc");
    AddStringToMsgPart(MsgHeader, VerStrHTML);
    AddStringToMsgPart(MsgHeader, "Patent%20Pending");
    AddStringToMsgPart(MsgHeader, "Field=0,Lane=1");
    {
        char msg[256];
        sprintf(msg,"Grid=%d",GridNumber);
        AddStringToMsgPart(MsgHeader, msg);
    }

    SetMessageWaiting(FALSE);
    QueueMsgToSend(&MsgHeader, NULL, NULL);

    sock_init();//J2H Check for re-calling
    Connect();

    //InitBenchLog();

    SetNextState(sBeginState);

    return Status;
sock_err:
    printf("error in sock_init() moment, code %d",Status);
    return 1;
} // end DoStateInit

I16 DoStateBeginState(void)
{
    static U32 retrycount = 0;
    I16 Status = StatusOK;
    TMasterTime BufTime;

    PDL("State: BeginState.  This is the \"Resting Ground\" for the state machine.  Any time a state-thread ENDS it should probably start-over here.");
    PDL("   Primarily housecleaning state that performs non-one-time initializations that do not require de-inits (no mallocs in this state, just zeroing out stuff perhaps)");
    PDL("   This state might do things like update-the-sensor-mask-registers, or similar.");

    ClearBlockCounters();
    ClearTimestamps();

    if (kbhit())  //--J2H
    {
        Rem("USER Exiting...");
        TERMINATE = StatusUserAbort;
    }

    if (s) sock_tick(s, &Status);

    if ( ! InPitch() )  //only check for messages and enet traffic when we're not busy with a pitch.
    {

        if (GetCurrentState() == sInit) return Status; //sock_tick produced a non-fatal error

        if (MsgToSend())
        {
            Deb("Sending Queued Message");
            sock_fastwrite(s, bufMsgToSend, *(U32 *)bufMsgToSend + 4);
/*
            char *pNextMsgPartStr = bufMsgToSend + 4;
            if ( *pU32(pNextMsgPartStr) )
            {
                Deb("Head,");
                pNextMsgPartStr += 4 + *pU32(pNextMsgPartStr);
            }
            else
            {
                Deb("!Head,");
                pNextMsgPartStr += 4;
            }
            if ( *pU32(pNextMsgPartStr) )
            {
                Deb(pNextMsgPartStr + 4);
                pNextMsgPartStr += 4 + *pU32(pNextMsgPartStr);
            }
            else
            {
                Deb("!Pitch");
                pNextMsgPartStr += 4;
            }
            if ( *pU32(pNextMsgPartStr) )
            {
                Deb(pNextMsgPartStr + 4);
                //pNextMsgPartStr += 4 + *pU32(pNextMsgPartStr);
            }
            else
            {
                Deb("!Err\r\n");
                //pNextMsgPartStr += 4;
            }
*/
            ClearMsgToSend();
            GetMasterTick(BufTime);
            LastKeepaliveBigTick = BufTime.Hi32;

        }

        GetMasterTick(BufTime);
        BufTime.Lo32 = BufTime.Hi32 - LastKeepaliveBigTick;
        if ( BufTime.Lo32 >= (BigTicksPerKeepalive) )
        {
            Rem("45 sec since last communication, issue keepalive.");
            QueueMsgToSend(NULL, NULL, NULL);
            LastKeepaliveBigTick = BufTime.Hi32;
        }

        SetNextState(sCheckAndParseMsg);
    }//end if inpitch
    else
    {
        SetNextState(sCheckAndReadSensors);
    }

    return Status;

sock_err:
    SetNextState(sInit);
    if ( MsgToSend() && ((*(U32 *)bufMsgToSend + 4) == 16) )
        printf("\rA TCP error %d occurred during keepalive. Reinitializing: %0LX",Status,retrycount++);
    else
        printf("\rA TCP error %d occurred after connect. Reinitializing: %0LX",Status,retrycount++);
    return(Status);
}


int hexbytetoi(char *str)
{
    U8 result = ((*str - '0') > 9) ? (*str - 'A' + 10) : (*str - '0');
    str++;
    result<<=4;
    result |= ((((*str - '0') > 9) ? (*str - 'A' + 10) : (*str - '0')));
    return result;
}

void parsemask(U8 which, char* s)
{
    I8 bytectr;
    U8 *ps = s;

    for(bytectr = 5; bytectr >= 0; bytectr--)
    {
        Mask.ByBeam.Beam[which][bytectr] = hexbytetoi(ps);
        ps += 2;
    }
}

/*

All expected incoming messages are formatted as a four-letter command,
followed by a direction character "=" or "?", followed by any needed
parameters for the specific command indicated.

The only currently implemented command is MASK, and only the WRITE
version ("=") is currently supported.

Another command expected is "SYNC", which will get/set the systemtime.

Another possible command would be "RSND", for "resend"
*/

void ParseAndExecuteMessage()
{
    //ClearDatums();
    U8 *ptr = bufMsgReceived + 4;//past length info
    U32 len = (*(U32 *)bufMsgReceived);

    //*(bufMsgReceived + len + 4) = '\0';//corrupt and ugly, deal with it, its just debug code until we add the parser
    //Deb(ptr);

    switch( *(U32 *)ptr )
    {
        case cmdSYNC:
            Rem("SYNC command received.");
            if ( len < 5 )
            {
                 Deb("SYNC but length is impossibly small. QueueResendForXmit()");
                 return;
            }
            ptr += 4;
            if (*ptr != '=')
            {
                Deb("SYNC but no = QueueResendForXmit()");
                return;
            }
            TMasterTime MTBuf;
            GetMasterTick(MTBuf);
            SyncOffsetTime = MasterTickToFloat(MTBuf);
            delay(100);
            MsgPitchData.Length = 0;
            AddStringToMsgPart(MsgPitchData, "100.0");
            AddStringToMsgPart(MsgPitchData, ",,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,");
            QueueMsgToSend(&MsgHeader, &MsgPitchData, NULL);
            ClearFlickerCounters();
        break;


        case cmdMASK:
            Rem("MASK command received.");

            ptr += 4;  PDL("skip 'MASK'");

            if ( *ptr == '=' )
            {
                if (len==5) //The MASK= command is empty, so clear masks
                {
                    ClearMask();
                    ClearFlickerCounters();
                }
                else //The MASK= command has data on it
                {
                    if ((len != 43) && (len != 31))
                    {
                        Deb("MASK but length is wrong.  QueueResendForXmit()");
                        return;
                    }
                    ptr ++;    PDL("skip '='");
                    if ((*(ptr+12) !=',') || (*(ptr+25) != ','))
                    {
                        Deb("'MASK=', but ',' missing.  QueueResendForXmit()");
                        return;
                    }

                    parsemask(0, ptr);
                    ptr += 13;
                    parsemask(1, ptr);
                    ptr += 13;
                    if (len == 43)
                        parsemask(2, ptr);
                    else
                    {
                        Rem("check to see if we got a short message despite being front grid?");
                    }
                    Rem("About to SetMask()");
                    SetMask();
                }//MASK= done being handled.
            }
            else if ( *ptr == '?' )
            {
                //ptr ++;    PDL("skip ?");
                BuildMaskFlickerErrorData();
                QueueMsgToSend(&MsgHeader, NULL, &MsgErrorData);
            }
            else
            {
                Deb("MASK but no = or ? QueueResendForXmit()");
                return;
            }
        break;


        case cmdPROG:
            Rem("PROG command received.");
            if ( len < 5 )
            {
                 Deb("PROG but length is impossibly small. QueueResendForXmit()");
                 return;
            }
            PDL("skip PROG");
            ptr += 4;
            len -= 4;

            if ( *ptr != '=' )
            {
                 Deb("PROG but no = QueueResendForXmit()");
                 return;
            }
            Print("PROG=, Spawning Updater");

            {
                char msg[8];
                char msg2[8];
                sprintf(msg,"%d",GridNumber);
                sprintf(msg2,"0x%04X",REVISION);
                DoStateDeInit();
                while ( spawnl(P_WAIT,"OPUPDATE.EXE",msg,msg,msg2,NULL) )
                {
                    printf("\nerrno == %d\n",errno);
                    delay(2000);
                }
                Reboot();
            }
        break;

        case cmdBOOT:
            Reboot();
        break;

        default:
             Deb(utod(*(U32 *)ptr));
             puts("unknown command received and attempted to parse.");
    }

    SetMessageWaiting(FALSE);
}

I16 DoStateCheckAndParseMsg(void)
{
    int Status = StatusOK;

    PDL("State: CheckAndParseMsg.  Determines if a enet message packet was received.  If so, calls the parser, which interrupts the entire state machine.");
    PDL("   This state simply branches the machine onto a different state-thread.");

    if ( MessageWaiting() )
    {
        ParseAndExecuteMessage();
    }

    SetNextState(sCheckAndReadSensors);
    return Status;
}

I16 DoStateCheckAndReadSensors(void)
{
    int Status = StatusOK;

    PDL("State: CheckAndReadSensors.  This state determines if any installed P104-DIO-96-S02 boards have detected a blocked, unmasked, sensor, and if they do, takes and stores data from the sensors in a temporary variable for initial analysis.");

    if ( GotData() )
    {
        if ( ! bInPitch )
        {
            SetInPitch(True);

        }
        asm CLI;
        GetDatum();
        asm STI;

        #if HARDWARE_VERBOSITY >= VERBOSITY_H_DEBUGREPORT
            char Buf[256] = "Datum: ";
            DatumToStr(MostRecentData.Datum, Buf);
            HDR(Buf);
        #endif

        if (AddDatum() == FALSE)
        {
            SetMask();
            ClearDatums();
        }

        SetNextState(sBeginState);
    }
    else
    {
        SetNextState(sNoDark);
    }

    return Status;
}

I16 DoStateNoDark(void)
{   char msg[256];
    int Status = StatusOK;

    PDL("State: NoDark.  This is the no-sensors-were-blocked state which branches based on history.  If there WAS data and NOW is not, it begins processing stuff.  Otherwise it just idles the machine.");

    if ( InPitch() )
    {
        SetInPitch(False);       //pitch or flicker has ended

        CheckFrameID("pitch");

        if ( NextRawData && (NextRawData <= RawDataCount) )
        {
            #if (SOURCE_VERBOSITY >= VERBOSITY_S_DEB)
                sprintf(msg,"Saw Pitch! %d Frames",NextRawData);
                Deb(msg);
            #endif

            SetNextState(sImplementCounters);
        }
        else
        {
            SetNextState(sBeginState);
        }
    }
    else
    {
        SetNextState(sBeginState);
    }

    return Status;
}

void DoIBeamCounters(U16 BeamIndex, double SampleTime, const TIBeam & IBeam)
{
    pU16 pScanWord = pU16(&IBeam);
    ForInt(IWord, 0, 2)
        if ( *pScanWord != 0xFFFF) //if any bit in word is "dark"
        {
            pU8 pScanByte = pU8(pScanWord);
            ForInt(IByte, 0, 1)
                if ( *pScanByte != 0xFF) //if any bit in this byte is "dark"
                {
                    U8 bMask = *pScanByte;
                    ForInt(IBit, 0, 7)
                        U16 MasterBitIndex = IWord * 16 + IByte * 8 + IBit;
                        if ( (bMask & 1) == 0 ) //if dark
                        {
                            if ( BlockCount.Counter[BeamIndex][MasterBitIndex] > 0 )
                            {
                                ++BlockCount.Counter[BeamIndex][MasterBitIndex];
                                pTTimestamps MyTimestamps = GetMyTimestamps(BeamIndex * 48 + MasterBitIndex);
                                MyTimestamps->EndTime = SampleTime;
                            }
                            else if ( BlockCount.Counter[BeamIndex][MasterBitIndex] == 0 )
                            {
                                BlockCount.Counter[BeamIndex][MasterBitIndex] = 1;
                                pTTimestamps MyTimestamps = GetMyTimestamps(BeamIndex * 48 + MasterBitIndex);
                                MyTimestamps->StartTime = SampleTime;
                                MyTimestamps->EndTime = SampleTime;
                            }
                            else if ( BlockCount.Counter[BeamIndex][MasterBitIndex] == -1 ) // if this bit is dark, but this bit counter was already "-1", then
                            {
                                TBD("Second intercept for this bit:");
                                TBD(utod(BeamIndex));
                                TBD(utod(MasterBitIndex));
                                BlockCount.Counter[BeamIndex][MasterBitIndex] = -2;
                            }
                        }
                        else  // if the byte had some dark, but this bit isn't dark,
                        {
                            if ( BlockCount.Counter[BeamIndex][MasterBitIndex] > 0 )//and this bit *was* dark
                            {
                                Rem("This test for \"is the blockage done\" doesn't catch the last bit in a byte.");
                                BlockCount.Counter[BeamIndex][MasterBitIndex] = -1;
                            }
                        }
                        bMask >>= 1;
                    EndFor;
                }
                ++pScanByte;
            EndFor;
        }
        ++pScanWord;
    EndFor;
}

I16 DoStateImplementCounters(void)
{
    int Status = StatusOK;

    PDL("State: ImplementCounters.  This state performs basic event counting on the ring buffered data.  (48*3) individual counters, one per sensor/laser pair.");
    PDL("If the counter gets to be almost RawDataCount, set that bit in the \"which lasers are blocked\" structure.");
    PDL("Also, make start and end times relative to start of the raw data, and convert start and end times to microseconds.");

    double FirstTime = MasterTickToFloat(RawData[0].SampleTime);
    ForInt(IFrame, 0, NextRawData - 1)
        DoIBeamCounters(0, (MasterTickToFloat(RawData[IFrame].SampleTime) - FirstTime) * USecsPerMasterTick, RawData[IFrame].X);
        DoIBeamCounters(1, (MasterTickToFloat(RawData[IFrame].SampleTime) - FirstTime) * USecsPerMasterTick, RawData[IFrame].Y);
        if ( DIO96Base[1] )
            DoIBeamCounters(2, (MasterTickToFloat(RawData[IFrame].SampleTime) - FirstTime) * USecsPerMasterTick, RawData[IFrame].LX)
        ;
    EndFor;

    ClearDatums();

    LastPitchTime = FirstTime - SyncOffsetTime;

    SetNextState(sParseData);

    return Status;
}

void CheckIBeamFlicker(TTimestamps IBeamResults[], I16 & NextResult)
{
    if ( NextResult )
    {
        SortTimestamps(IBeamResults, NextResult, byLength);
        U32 dT = IBeamResults[0].EndTime - IBeamResults[0].StartTime;
        if ( dT < MIN_REAL_LONGEST )
        {
            #if ( SOURCE_VERBOSITY >= VERBOSITY_S_DEB )
                char Buf[256];
                sprintf(Buf, "It was just flicker, longest intercept %lfus.", double(dT));
                Deb(Buf);
            #endif
            ForInt(I, 0, NextResult - 1)
                CountFlicker(IBeamResults[I].BitIndex);
            EndFor;
            NextResult = 0;
        }
        else if ( NextResult > MAX_REAL_LASERS_INTERCEPTED )
        {
            Rem("Call the rest flicker.");
            ForInt(I, MAX_REAL_LASERS_INTERCEPTED, NextResult - 1)
                CountFlicker(IBeamResults[I].BitIndex);
            EndFor;
            NextResult = MAX_REAL_LASERS_INTERCEPTED;
        }
    }
}

I16 DoStateParseData(void)
{
    int Status = StatusOK;

    PDL("State: ParseData.  If there was data and now is not, here we parse it all out as t1-tn (and probably detect flicker).");

    {
        CheckIBeamFlicker(XResults, NextXResult);
        CheckIBeamFlicker(YResults, NextYResult);
        CheckIBeamFlicker(LXResults, NextLXResult);

        if ( NextXResult )
        {
            if ( NextLXResult )
            {
                Rem("Both X and LX...");
                if ( XResults[0].EndTime - XResults[0].StartTime > LXResults[0].EndTime - LXResults[0].StartTime )
                {
                    Rem("X was bigger, call LX flicker.");
                    ForInt(I, 0, NextLXResult - 1)
                        CountFlicker(LXResults[I].BitIndex);
                    EndFor;
                    NextLXResult = 0;
                    Status = StatusNopeJustFlicker;
                }
                else
                {
                    Rem("LX was bigger, call X flicker.");
                    ForInt(I, 0, NextXResult - 1)
                        CountFlicker(XResults[I].BitIndex);
                    EndFor;
                    NextXResult = 0;
                    Status = StatusNopeJustFlicker;
                }
            }
            if ( NextYResult == 0 )
            {
                Rem("No Y; flicker or other lane.");
                Status = StatusNopeJustFlicker;
            }
        }
        else
        {
            if ( NextLXResult )
            {
                if ( NextYResult == 0 )
                {
                    Rem("No Y; flicker or other lane.");
                    Status = StatusNopeJustFlicker;
                }
            }
            else
            {
                Rem("No X or LX; flicker or other lane.");
                Status = StatusNopeJustFlicker;
            }
        }
    }

    #if ( SOURCE_VERBOSITY >= VERBOSITY_S_DEB )
        char Buf[256];
        if ( NextXResult )
        {
            sprintf(Buf, "X longest intercept %lfus.", double(XResults[0].EndTime - XResults[0].StartTime));
            Deb(Buf);
        }
        if ( NextLXResult )
        {
            sprintf(Buf, "LX longest intercept %lfus.", double(LXResults[0].EndTime - LXResults[0].StartTime));
            Deb(Buf);
        }
        if ( NextYResult )
        {
            sprintf(Buf, "Y longest intercept %lfus.", double(YResults[0].EndTime - YResults[0].StartTime));
            Deb(Buf);
        }
    #endif

    if ( Status == StatusOK )
        SetNextState(sQueueDataForXmit);
    else
        SetNextState(sBeginState)
    ;

    return Status;
}

void AddIBeamResultsToPitchData(char * BeamName, TTimestamps (& BeamResults)[], I16 & NextBeamResult)
{
    if ( NextBeamResult )
    {
        AddStringToMsgPart(MsgPitchData, BeamName);
/*
        ForInt(I, 0, NextBeamResult - 1)
            AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].BitIndex % 48));
            AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].StartTime));
            AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].EndTime));
        EndFor;
*/
        ForInt(I, 0, 2)
            if (I<NextBeamResult)
            {
               AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].BitIndex % 48));
               AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].StartTime));
               AddStringToMsgPart(MsgPitchData, utod(BeamResults[I].EndTime));
            }
            else
               AddStringToMsgPart(MsgPitchData, ",,");
        EndFor;
    }
}

I16 DoStateQueueDataForXmit(void)
{
    int Status = StatusOK;
    char Buf[256];

    PDL("State: QueueDataforXmit.  Once we've arrived at t1-tn for each axis we build a packet for ethernet transmission here.  This data also includes our flicker counters or similar.");

    MsgPitchData.Length = 0;
    sprintf(Buf, "%.1lf", double(LastPitchTime * MSecsPerMasterTick));
    AddStringToMsgPart(MsgPitchData, Buf);
    AddIBeamResultsToPitchData("X,0", XResults, NextXResult);
    AddIBeamResultsToPitchData("Y,1", YResults, NextYResult);
    AddIBeamResultsToPitchData("LX,2", LXResults, NextLXResult);
    ClearTimestamps();
    BuildBlockFlickerErrorData();
    //MsgErrorData.Length = 0;
    //AddStringToMsgPart(MsgErrorData, ",,"); //Skip block data.
    //AddStringToMsgPart(MsgErrorData, FlickersToStr());
    QueueMsgToSend(&MsgHeader, &MsgPitchData, &MsgErrorData);
    delay(200);
    SetNextState(sBeginState);

    return Status;
}

I16 DoStateQueueErrorForXmit(void)
{
    int Status = StatusOK;

    PDL("State: QueueErrorForXmit.  Basically if we determine flicker is bad enough we queue the data into part of a enet packet for sending in this state.");

    BuildBlockFlickerErrorData();
    //MsgErrorData.Length = 0;
    //AddStringToMsgPart(MsgErrorData, ",,"); //Skip block data.
    //AddStringToMsgPart(MsgErrorData, FlickersToStr());
    QueueMsgToSend(&MsgHeader, NULL, &MsgErrorData);

    SetNextState(sBeginState);

    return Status;
}

I16 DoStateDeInit(void)
{
    int Status = StatusOK;

    PDL("State: DeInit.  Perform any one-time de-initialization needed here.  Consider moving out of state machine.  This *must* be the counterfunction to StateInit.");

    if (s)
    {
        sock_close( s );
        //sock_wait_closed( (sock_type *)s, sock_delay, NULL, &status );
    }

    //DeInitBenchLog();

    sock_exit();
    return TERMINATE;

    sock_err:
        printf("\nAborting@@@!!! Code %d",Status);
        return(Status);
}

//
// End State Logic
//**************************************************************************************************************************************************

//**************************************************************************************************************************************************
// Begin Main Program Logic
//

int MainProgramLoop(void)
{
    PDL("int MainProgramLoop(void): contains the main \"while not done\" loop in which is everything other than init and deinit, theoretically");
    U16 Status;

    PDL("State Machine Entry Point and Main Program Loop.  Loop until TERMINATE is not StatusOK");
    while ( TERMINATE == StatusOK )
    {
        switch ( GetCurrentState() )
        {
            case sInit:
                PDL("STATE_INIT - should happen only one time, but could be done if TCP pipe is ever closed.  Program kinda starts over by going to this state again.");
                Status = DoStateInit();
            break;
            case sBeginState:
                Status = DoStateBeginState();
            break;
            case sCheckAndParseMsg:
                Status = DoStateCheckAndParseMsg();
            break;
            case sCheckAndReadSensors:
                Status = DoStateCheckAndReadSensors();
            break;
            case sImplementCounters:
                Status = DoStateImplementCounters();
            break;
            case sNoDark:
                Status = DoStateNoDark();
            break;
            case sParseData:
                Status = DoStateParseData();
            break;
            case sQueueDataForXmit:
                Status = DoStateQueueDataForXmit();
            break;
            case sQueueErrorForXmit:
                Status = DoStateQueueErrorForXmit();
            break;
            case sDeInit:
                PDL("STATE_DEINIT - should happen only one time, consider moving into a DEINIT function instead of inside the state machine -- J2H");
                Status = DoStateDeInit();
            break;
            default:
                 TERMINATE = StatusFatalError;
                 Deb("FATAL_ERROR:State Switch Default Reached.  This is bad.  State WAS:");
                 Deb(GetStateName[GetPreviousState()]);
            //break;

        } End("Switch state machine");

        ChangeState();

    }//end while  //end of main program loop

    return Status;
}

#pragma argsused
int main(I16 ParamCount, char* ParamStr[])
{
    char msg[256];

    PDL("Initialize the screen and print hello world type info");
    textmode(C4350);
    directvideo=TRUE;
    clrscr();

    if ( ParamCount <= 1 )
    {
        PDL("Print instructions on how to run the program correctly");
        Print("Run with 'LIVEMODE gridnumber' \n");
        return 0;
    }
    else
    {

        TBD("Process command-line parameters.");
        GridNumber = atoi(ParamStr[1]);
    }

    Print("OnePitch Livemode "VerStr" by ACCES on " __DATE__"\r\n"); //--J2H commented because I don't have a verinfo structure etc.


#ifdef InitHTTPResponder
    InitHTTPResponder();
#endif

    MainProgramLoop();
    if(s) sock_close(s);
#ifdef h
    if(h) sock_close(h);
#endif

    return 0;
}

//
// End State Logic
//**************************************************************************************************************************************************

/*  Questions that need answers

Q1: What units are we going to return t1-t6 in?
A1: usec

Q2: What is the data format in the packet sent? (Are we going to return t1-t6, or t1[axis]-tn[axis] where "n" is the maximum number of blocked sensors EVER to squeeze into one udp packet (always return data even if tn[a]==undef) etc)
A2: note: all lengths are exclusive
    U32 PacketLength
    U32 HeaderLength
        CSV Header

    U32 PitchLength
        CSV Pitch, with the following defined fields
        timestamp in ms.
        array[0..n] of                        //n can be 1 or 2
          string which "X,"Y","LX"
          number which (0,1,2)
          ary[0..numinterceptsonthatibeam]
            lasernumberonthatibeam
            intercept time (rectified)
            de-intercept   (rectified)
          endary
        endary

    U32 ErrorLength
        CSV Errordata, with the following defined fields
        blocks on X,
        blocks on Y,
        blocks on LX,
        flickercounters on X,
        flickercounters on Y,
        flickercounters on LX

    No reply is necessary to this message packet.
    Any combination of fields can be blank.
    Empty fields in CSV portions are indicated by ",,"

    If a message sent includes BLOCK bits, it is expected but not required
    that the boss-comp will issue a MASK= message/response.
    The format of MASK= is as follows:
*/

/*  Revision History (informal)
Revision 0.01: Started framework implementation for robust Debug modes. - J2H
Revision 0.02: Modified original framework to improve Debugging, such as switching #defines to typedefs where useful. - DW
Revision 0.03: Added VERBOSITY_S_SILENT, and implemented a "Deb" level of remark, more changes in this area
               Modified main() to echo the long-term intent (by implementing "MainProgramLoop()", basically.
               proto-defined the state enum and some related code - J2H
Revision 0.04: Coded fast itod(), enum Bool
               Changed a few ">= silent"s to "> silent"s - DW
Revision 0.05: Coded utod(), imported INSW()/InPortDW()/OutPortDW(), fixed up motorola byte-order functions - DW
Revision 0.06: Changed some variable/parameter declarations to use our typedefs.  Began integrating "HandleDiscovery()" function - J2H
Revision 0.07: Added GOTDATA - DW
Revision 0.08: Added three new versions of GOTDATA.  Discuss.  Implemented Printer Byte Write - J2H
Revision 0.09: Added state-machine functions, to support nested state machines later - DW
Revision 0.0A: Added PDL to each state function.  Continued main().  Tried to get it to compile. - J2H
Revision 0.0B: Added main data structures, added a couple main data functions, added PCIDetect(). - DW
Revision 0.0C: Commented.  Moved GetDatum into Hardware Access region.  - J2H
Revision 0.0D: Worked with Dare on getting it to a minimalist running condition, added string conversion INC, filled in some "implement  Counters" code.
Revision 0.0E: found sock_puts(t,msg), the important TCP/IP stuff. -J2H
Revision 0.10: Many changes. -DW and J2H
Revision 0.11: implementing connect() and message queuing, both in and out.  -J2H and DW
Revision 0.12-0.19.  Many changes lol.  J2H DW.
Revision 0.20: First "valid" Pitch.  12/1/2008
Revision 0.21: Increased block buffer size to allow REALLY slow balls without hitting the block line
Revision 0.23: Added "PROG= and related commands.
Revision 0.25: fixed hangup associated with "Cleave"
Revision 0.26: improved(maybe fixed) timing
Revision 0.27: fixed CPU speed display, improved upload process
Revision 0.28: fault tolerance in upload process
Revision 0.29: Moved sock_tick into if ! InPitch.
Revision 0.30: Fixed rad vs diameter issue in flicker rejection, added DONE=length checking;
Revision 0.31: Added debug in search of block loop error.
Revision 0.32: Reduced min real intercept to still report when longest intercept is on a blocked laser.
Revision 0.33: Replaced PROG= to incorporate OPUpdate.exe functionality, fixed a typo in the first print statement, and resolved a double-block on startup
Revision 0.34: Changed to use sock_exit before system(), added reply to SYNC=.
Revision 0.35: Changed AddDatum to clear bInPitch /on/ the block tick, not just past it, because we won't go past it if it clears at just the right moment. Changed getch()es to ComplainUntilKBHit()s so they can hear when it happens.
Revision 0.36: Added blocks to pitch error string.
Revision 0.37: Removed blocks from pitch error string because they were apparently causing synch errors, and added MASK? command.
Revision 0.38: Debug build to benchmark sock_tick, fix BuildMaskFlickersErrorData ForInt problem
Revision 0.39: Debug build bandaid for fast grid 3 / long sock_tick. 300ms
Revision 0.40: Modified ReportBlocked to only pay attention to the last half of the buffer in order to handle "cleaves" as blocks, and refactored GetMyTimestamps to be per-axis in the first place.
Revision 0.41: Modified HandleTCP to re-init on length errors, and sock_exit in state_init
Revision 0.42: Removed sock_exit (was locking up), moved sock_tic(s) back out of !inpitch
Revision 0.43: Removed sock_exit (was locking up), moved sock_tic(s) back out of !inpitch
Revision 0.44: Reserved as a test for the upload process
Revision 0.45: Changed "system" to "spawnl" in prog=
Revision 0.46: Added a sock_tick(s) to xmit code, and a print to handletcp
Revision 0.47: Auto-MASKOKOK.
Revision 0.48: Clear flicker counters when "SYNC=" or "MASK="(but not "MASK=...") received.
Revision 0.49: Added blocks back in whenever error data is sent.
*/

/* Features pulled due to time constraints

DONE: 1) upload new program across ethernet via TCP commands
7) basic http TCP_Listen on port 80, with config file stuff supported via forms
6) handle miscellaneous config file issues (currently there are many unsupported possibilities defined)
5) log all ethernet packets forever
DONE: 4) handle command-line parameters (none planned, but could be)
8) uptime and msg ID fields in the header
3) report no P104-DIO-96 cards detected, and related errors (second card unverifiable, etc)
DONE: 2) handle condition where flicker occurred during a pitch on an adjacent sensor better (currently thrown away) (second intercept)
DONE:   add more bits to timestamp (not wrap at 4 seconds in header)
DONE:   support MASK? (MASKWHAT)
DONE:   implement "BOOT=" command which reboots the system immediately


*/

/*  Master List of global variables loaded from config file ()

I   D   Description       (I = implemented, D = intended to be documented)

N - N - PCI Card #0 base address override
N - N - PCI Card #1 base address override
N - Y - TCP/IP Timeout period
N - Y - Boss-Comp Connection retries
        This box's IP address
        BOOTP/DHCP mode
        port number on boss-comp to connect to
        DNS "name" of boss-comp (or IP address) for "resolve"
        initial hardware MASK state
        instance location data (gps coords, installation name, owner, creator, installer, team, coach, serial number, ad nauseum)
        ACCES copyright information
        OnePitch copyright information
        Applicable Patent numbers
        debug mode to run in (other than hardware-level which is purely compile time)
        where to direct debug strings (com1, enet, console, etc) and any config info like IP, Port, Address, etc.
        http listen port (default is 80)
        RawDataCount
        MAX_MPH
        P104-DIO-96 VID & PID
        # of lanes
        position information for lasers (offset from ground, center, space between lasers, etc etc)
*/


/*
TBD:
  logo.sys custom graphic -- need name of product
  debug why flicker counters don't look like we think they oughta, although they look better now

TBD someday:
  #define these functions to ease the math-typos
    ConvertTicksToUSecs()
    ConvertTicksToSecs()
  half-done:implement "5 minute TCP retry" and "1 hour TCP reboot" connection timers
  implement an actual "queue" for queuemsgtosend() so multiple outgoing messages can be waiting to be sent (right now it only sends the last one "queued")





*/